<?php
namespace Folha\Logic\ESocial;

use \Folha\Logic\Esocial\Validator\BasicResult;
use \Folha\Logic\Esocial\Validator\References\ValorCampo;
use \Folha\Logic\Esocial\Validator\References\ForeignKey;
use \Folha\Logic\Esocial\Validator\References\Label;
use \Folha\Logic\Esocial\Validator\References\FixedField;
use \Folha\Logic\Esocial\Validator\References\FixedValue;
use \Folha\Logic\Esocial\Validator\References\Executavel;
use \Folha\Logic\Esocial\Validator\References\PadraoInvalido;

/**
 * Constroi o conjunto de dados que sera retornado como resultado da validação dos eventos.
 *
 * Class DataValidator
 * @package Folha\Logic\ESocial
 */
class DataValidator
{
	private $estrutura;
	private $data;
	private $eventos;
	private $rules;
	private $fixedFields;
    private $excludedTables;
	private $padroesMetadados;
	private $resultado;

    static $pk = array();

    /**
     *
     *
     * @param $estrutura  array  Relacao dos campos dos eventos do E-Social que serão validados (esturura.json).
     * @param $data       array  Array com todos os dados gerados.
     * @param $helpers    array  Array com todos as regras de validação dos campos dos eventos e dados auxiliares da validação. (helper.json)
     *                           O arquivo helper deve possuir as regras (rules) de cada evento e um array "campos" com os campos descritos na estrutura. O primeiro campo
     *                           deve ser um ID para identificar o registro.
     *                           Caso o campo possua filhos ele deve ser descrito com o seu nome conforme a geração e deve ser um array com as chaves sendo os nomes
     *                           dos campos filhos. Caso necessário também pode ter um campo ID para identificação na exibição do erro e cada campo filho, com excessão do ID,
     *                           deve possuir dentro das regra uma chave "array" com o nome do campo pai.
     */
    public function __construct($estrutura, $data, $helpers)
	{
		$this->estrutura      = $estrutura;
		$this->data           = $data;
		$this->rules          = $helpers["rules"];
        $this->fixedFields    = $helpers["fixedFields"];
        // Armazena o nome dos eventos (presentes no arquivo estrutura.json).
        $this->eventos        = array();
        // Objeto onde todas as entidades criadas serao planificadas. Este array sera retornado no final do processo de geracao dos dados.
        $this->resultado      = array();

        $this->excludedTables = $helpers["excludedTables"];

        // Este array armazena todos os padroes que sao compreendidos pelo modelo, e sao utilizados nos arquivos de estrutura.json e helper.json. Qualquer classe
        // utilizada neste array deverá implementar a interface de \Folha\Logic\Esocial\Validator\References\Reference.
		$this->padroesMetadados = array(
            new ValorCampo(),
            new ForeignKey(),
            new Label(),
            new FixedField(),
            new FixedValue(),
            new Executavel(),
            new PadraoInvalido(),
		);
	}

	public function addEvento($evento)
	{
		$this->eventos[] = $evento;
	}

	public function build() 
	{
        $this->percorreLinhasDados();
		return $this->resultado;
	}

    /**
     * Percorre todas as linhas do evento.
     */
    protected function percorreLinhasDados()
	{
        foreach ($this->estrutura as $nomeEvento => $campos) {
            if (in_array($nomeEvento, $this->eventos)) {
                $resultContainer = array();
                foreach ($this->data as $key => $dados) {
                    $result = new BasicResult($key);
                    $this->percorreEventos($nomeEvento, $dados, $result);
                    array_push($resultContainer, $result);
                    unset($this->data[$key]);
                }
                $this->resultado[$nomeEvento] = $resultContainer;
            }
        }
	}

    /**
     * Percorre todos os agrupamentos do arquivo estrutura.json
     *
     * @param $nomeEvento
     * @param $linhaDados array
     * @param $result
     * Passa por este metodo somente para ser transmitido para os proximos, onde este realmente sera necessario.
     */
    protected function percorreEventos($nomeEvento, $linhaDados, BasicResult $result)
	{
        if (isset($this->rules[$nomeEvento])) {
            $this->executaRegrasDeEvento($this->rules[$nomeEvento], $nomeEvento, $linhaDados, $result);
        }
	}

    /**
     * Percorre todas as regras descritas para determinada tabela (se estas existirem). Apos executar todas as regras desta tabela,
     * este metodo tambem faz chamada ao metodo que executa as regras para os campos destas tabelas.
     *
     * As regras sao descritas no arquivo helper.json, sendo que cada regra deve conter as chaves de "condicao", onde é descrita
     * a condicao a ser verificada, "entao", onde e descrito o que deve ser feito, caso a condicao seja satisfeita, e "senao", que
     * descreve o que deve ser feito caso a condicao nao seja satisfeita.
     *
     * Importante notar que, sempre que a condicao resultar em "false", a entidade (tabela) em questao sera removida do result.
     *
     * @param $evento                          array Array contendo o conjunto condica, entao, senao
     * @param $nomeEvento
     * @param BasicResult $result              Container armazenando todas as entidades de determinada linha.
     * @param $linhaDados                      array Array contendo os dados da linha do arquivo excel.
     */
    protected function executaRegrasDeEvento($evento, $nomeEvento, $linhaDados, BasicResult $result)
	{
        foreach ($evento["regras"] as $regra) {
            $condicao = $this->verificaCondicaoRegra(
                $regra["condicao"], null, $result, $linhaDados
            );
            if ($condicao) {
                $this->getDadosReference(
                    null, null, $regra["entao"], null, $result, $linhaDados
                );
            } else {
                $result->removeResult(null);
                if (isset($regra["senao"])) {
                    $this->getDadosReference(
                        null, null, $regra["senao"], null, $result, $linhaDados
                    );
                }
            }
        }

        if (isset($evento["campos"])) {
            $fields = $this->executaRegrasDeCampo($evento["campos"], $nomeEvento, $linhaDados, $result);
            $result->setFields($fields);
        }
	}

    /**
     * Executa as regras a nivel de coluna. Assim como nas regras de tabela, estas devem possuir condicao, entao e senao.
     *
     * @param $regrasColuna
     * @param $nomeTabela
     * @param BasicResult $result
     * @param $linhaDados
     * @return bool
     */
    private function executaRegrasDeCampo($campos, $nomeEvento, $linhaDados, BasicResult $result)
    {
        $arrayErros = array();
        foreach ($campos as $nomeCampo => $campo) {
            $errosCampo = array();

            if (!array_key_exists("regras", $campo) && is_array($linhaDados[$nomeCampo])) {
                $errosFilhos = array();
                foreach ($linhaDados[$nomeCampo] as $id=>$dados) {
                    $errosFilhos[$id] = $this->executaRegrasDeCampo($campo, $nomeEvento, $dados, $result);
                }
                //array_push($errosCampo, $errosFilhos);
                $errosCampo = $errosFilhos;
            }
            else {
                foreach ($campo['regras'] as $regra) {
                    $condicao = $this->verificaCondicaoRegra(
                        $regra["condicao"], $nomeEvento, $result, $linhaDados
                    );

                    if ($condicao) {
                        $resultado = $this->getDadosReference(
                            $nomeEvento, null, $regra["entao"], null, $result, $linhaDados
                        );
                        array_push($errosCampo, $resultado);
                        //$errosCampo = $resultado;
                    } else if (isset($regra["senao"])) {
                        $resultado = $this->getDadosReference(
                            $nomeEvento, null, $regra["senao"], null, $result, $linhaDados
                        );
                        //array_push($errosCampo, $resultado);
                        $errosCampo = $resultado;
                    }
                }
            }

            $arrayErros[$nomeCampo] = $errosCampo;
        }
        return $arrayErros;
    }

    /**
     * Metodo utilizado para resolucao da condicao.
     *
     * @param $regra          string            Padrao informado no helper.json, que apos processado, possui a condicao necessaria.
     * @param $nomeTabela     string            Nome da tabela sobre a qual a condicao sera imposta
     * @param BasicResult $result  Container com todas as entidades de determinada linha.
     * @param $linhaDados     array             Linha de dados
     *
     * @return boolean
     */
    protected function verificaCondicaoRegra($regra, $nomeEvento, BasicResult $basicResult, $linhaDados)
	{
		$condicao = $this->getDadosReference($nomeEvento, null, $regra, null, $basicResult, $linhaDados);
        $result = null;
		eval('$result = ('.$condicao.');');
		return $result;
	}


	// Retorna o valor esperado para uma coluna informada nos agrupamentos através de meta-dados.
	protected function getDadosReference($nomeEvento, $nomeColuna, $metaDado, $valorFixo,
                                         BasicResult $basicResult, $linhaDados = null)
	{
		$padraoEncontrado = false;
        $matches = array();
        /** @var $padrao \Folha\Logic\Esocial\Validator\References\Reference */
		foreach ($this->padroesMetadados as $padrao) {
			if (preg_match_all($padrao->getRegularExpression(), $metaDado, $matches)) {
                /* Itera a ocorrencia de varios padroes juntos */
                for ($i = count($matches[0]); $i > 0; $i--) {
                    $metaDado = $padrao->getValorReferencia(
                        $nomeEvento, $nomeColuna, $metaDado, $valorFixo, $basicResult, $linhaDados, $this->fixedFields
                    );
                }
				$padraoEncontrado = true;
			}
		}
        
		return !$padraoEncontrado ? $metaDado : $this->getDadosReference (
            $nomeEvento, $nomeColuna, $metaDado, $valorFixo, $basicResult, $linhaDados
		);
	}
}